from typing import Tuple


class Task:

    def __init__(self, id_: str, name: str, samples: list[Tuple[str, str]]):
        self.id_ = id_
        self.name = name
        self.samples = samples


class Parser:

    def __init__(self, tasks: list[Task]):
        self.tasks = tasks



import importlib.metadata
import os
from typing import Any, Optional

import click

from lucy import update_snippets as us
from lucy.config import Config, SampleType, Website
from lucy.dbx_client import DropboxClient
from lucy.filesystem import LocalFS
from lucy.parser_.contest import ContestParser
from lucy.tester import Tester


@click.group()
@click.version_option(importlib.metadata.version('lucy01'))
@click.pass_context
def cli(_: Any) -> None:
    LocalFS.setup()


@cli.command('update-snippets')
@click.option('-ed',
              '--entry-dir',
              'entry_dir_',
              default=Config.COMMONS_DIR,
              type=click.Path(exists=True))
@click.option('-o', '--out', 'out', default=Config.SNIPPETS_PATH, type=click.Path())
def update_snippets(entry_dir_: str, out: str) -> None:
    entry_dir_ = os.path.abspath(entry_dir_)
    out = os.path.abspath(out)
    snippets = us.run(entry_dir_, out)
    click.secho(f'Found {len(snippets)} snippets.', fg='green', bold=True)
    for snippet in snippets:
        click.echo(snippet)


@cli.command('setup')
@click.argument('site', type=Config.CLI_WEBSITE_CHOICE)
@click.argument('contest_id')
@click.argument('task_id', required=False, default=None, type=str)
@click.argument('test_id', required=False, default=None, type=str)
def setup(site: str, contest_id: str, task_id: Optional[str], test_id: Optional[str]) -> None:
    website = Website.from_string(site)
    if test_id is not None:
        assert task_id is not None
        if website != Website.ATCODER:
            raise NotImplementedError()
        in_path, out_path, idx = LocalFS.get_new_sample_paths(website, contest_id, task_id)
        DropboxClient().download(f'/{contest_id}/{task_id}/{SampleType.IN}/{test_id}', in_path)
        DropboxClient().download(f'/{contest_id}/{task_id}/{SampleType.OUT}/{test_id}', out_path)
        click.secho(f'Stored sample {test_id} as Test#{idx:02d}.', fg='green', bold=True)
        return
    contest_ = ContestParser(website, contest_id, task_id)
    for task in contest_.parser.tasks:
        LocalFS.store_samples(website, contest_id, task)
        LocalFS.create_impl_file(website, contest_id, task.id_)


@cli.command('test')
@click.argument('site', type=Config.CLI_WEBSITE_CHOICE)
@click.argument('contest_id')
@click.argument('task_id')
@click.argument('test_id', default=None, type=int, required=False)
@click.option('-v', '--verbose', 'verbose', is_flag=True, default=False)
def test(site: str, contest_id: str, task_id: str, test_id: Optional[int], verbose: bool) -> None:
    website = Website.from_string(site)
    tester = Tester(website, contest_id, task_id, test_id)
    tester.run(verbose)



import dropbox  # type: ignore[import-untyped]

from lucy.config import Config


# pylint: disable=too-few-public-methods
class DropboxClient:

    def __init__(self, link: str = Config.ATCODER_TESTCASES_DROPBOX_LINK):
        self.shared_link = dropbox.files.SharedLink(link)

        token = Config.DROPBOX_TOKEN
        self.dbx = dropbox.Dropbox(token)

    def download(self, path: str, dest: str) -> None:
        self.dbx.sharing_get_shared_link_file_to_file(dest, self.shared_link.url, path)



from __future__ import annotations
import subprocess
from typing import Optional, Tuple

import click

from lucy.config import Website
from lucy.filesystem import LocalFS


# pylint: disable=too-few-public-methods
class Tester:

    def __init__(self, website: Website, contest_id: str, task_id: str,
                 test_id: Optional[int]) -> None:
        self.website = website
        self.contest_id = contest_id
        self.task_id = task_id

        self.num_tests = LocalFS.num_samples(website, contest_id, task_id)

        self.test_ids: list[int] = list(range(self.num_tests))
        if test_id is not None:
            if test_id < 0 or test_id >= self.num_tests:
                raise ValueError("Invalid `test_id`")
            self.test_ids = [test_id]

        self.impl_path = LocalFS.get_impl_path(website, contest_id, task_id)
        self.bin_path = LocalFS.get_impl_path(website, contest_id, task_id, bin_=True)

    def __compile(self) -> None:
        subprocess.check_call(['g++', self.impl_path, '-o', self.bin_path])

    def __exec(self, in_txt: str, truth_txt: str) -> Tuple[bool, str]:
        with subprocess.Popen([self.bin_path], stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE) as process:
            assert process.stdin
            process.stdin.write(in_txt.encode())
            process.stdin.close()
            assert process.stdout
            out_txt = process.stdout.read().decode()
            process.wait()
        return out_txt.strip() == truth_txt.strip(), out_txt

    def run(self, verbose: bool = False) -> None:
        self.__compile()

        for i, (in_txt, truth_txt) in zip(
                self.test_ids,
                LocalFS.tests(self.website, self.contest_id, self.task_id, self.test_ids)):
            click.echo(f'Test#{i:02d}', nl=False)
            click.echo(f'{"." * 50}', nl=False)
            passes, out_txt = self.__exec(in_txt, truth_txt)
            if passes:
                click.secho('AC', bg='green')
            else:
                click.secho('WA', bg='red')
                if verbose:
                    click.secho("Input:", bg='white', bold=True)
                    print(in_txt.strip())
                    click.secho("Output:", bg='red', bold=True)
                    print(out_txt.strip())
                    click.secho("Expected:", bg='green', bold=True)
                    print(truth_txt.strip())



import os
import shutil
from typing import Generator, Optional, Tuple

import click

from lucy.config import Config, SampleType, Website
from lucy.parser_.parser_ import Task


class LocalFS:

    @staticmethod
    def get_impl_path(website: Website,
                      contest_id: str,
                      task_id: str,
                      dir_: bool = False,
                      bin_: bool = False) -> str:
        impl_path = f'{Config.LUCY_HOME}/{website}/{contest_id.upper()}/{task_id.upper()}'
        if bin_:
            return f'{impl_path}/{Config.IMPL_BIN}'
        if not dir_:
            impl_path = f'{impl_path}/{Config.IMPL_MAIN}'
        return impl_path

    @staticmethod
    def get_sample_path(website: Website,
                        contest_id: str,
                        task_id: str,
                        type_: Optional[SampleType] = None,
                        idx: Optional[int] = None) -> str:
        sample_path = LocalFS.get_impl_path(website, contest_id, task_id, dir_=True)
        sample_path += f'/{Config.SAMPLES_DIR}'
        if type_:
            sample_path += f'/{type_}'
            if idx is not None:
                sample_path += f'/{idx:02d}.txt'
        return sample_path

    @staticmethod
    def delete_samples(website: Website, contest_id: str, task_id: str) -> bool:
        samples_dir = LocalFS.get_sample_path(website, contest_id, task_id)
        if os.path.exists(samples_dir):
            shutil.rmtree(samples_dir)
            return False
        return True

    @staticmethod
    def num_samples(website: Website, contest_id: str, task_id: str) -> int:
        return len(os.listdir(LocalFS.get_sample_path(website, contest_id, task_id, SampleType.IN)))

    @staticmethod
    def get_new_sample_paths(website: Website, contest_id: str,
                             task_id: str) -> Tuple[str, str, int]:
        idx = LocalFS.num_samples(website, contest_id, task_id)
        in_path = LocalFS.get_sample_path(website, contest_id, task_id, SampleType.IN, idx)
        out_path = LocalFS.get_sample_path(website, contest_id, task_id, SampleType.OUT, idx)
        return in_path, out_path, idx

    @staticmethod
    def store_sample(website: Website, contest_id: str, task_id: str, sample: Tuple[str,
                                                                                    str]) -> int:
        input_, output = sample
        in_path, out_path, idx = LocalFS.get_new_sample_paths(website, contest_id, task_id)

        with open(in_path, 'w+', encoding='utf-8') as f:
            f.write(input_)
        with open(out_path, 'w+', encoding='utf-8') as f:
            f.write(output)

        return idx

    @staticmethod
    def store_samples(website: Website, contest_id: str, task: Task) -> int:
        LocalFS.delete_samples(website, contest_id, task.id_)
        os.makedirs(LocalFS.get_sample_path(website, contest_id, task.id_, SampleType.IN),
                    exist_ok=True)
        os.makedirs(LocalFS.get_sample_path(website, contest_id, task.id_, SampleType.OUT),
                    exist_ok=True)
        for sample in task.samples:
            LocalFS.store_sample(website, contest_id, task.id_, sample)
        return len(task.samples)

    @staticmethod
    def setup() -> None:
        if not os.path.exists(Config.LUCY_HOME):
            click.echo(f'Creating LUCY_HOME ({Config.LUCY_HOME}).')
            os.makedirs(Config.LUCY_HOME, exist_ok=True)

        if not os.path.exists(Config.SNIPPETS_DIR):
            click.echo(f'Creating SNIPPETS_DIR ({Config.SNIPPETS_DIR}).')
            os.makedirs(Config.SNIPPETS_DIR, exist_ok=True)

        if not os.path.exists(Config.COMMONS_DIR):
            click.echo(f'Creating COMMONS_DIR ({Config.COMMONS_DIR}).')
            os.makedirs(Config.COMMONS_DIR, exist_ok=True)

        if not os.path.exists(Config.TEMPLATE_PATH):
            with open(Config.TEMPLATE_PATH, 'w+', encoding='utf-8'):
                pass

    @staticmethod
    def create_impl_file(website: Website, contest_id: str, task_id: str) -> str:
        impl_path = LocalFS.get_impl_path(website, contest_id, task_id)
        if not os.path.exists(impl_path):
            shutil.copy(Config.TEMPLATE_PATH, impl_path)
        return impl_path

    @staticmethod
    def read(path: str) -> str:
        with open(path) as f:  # pylint: disable=unspecified-encoding
            return f.read()

    @staticmethod
    def tests(website: Website, contest_id: str, task_id: str,
              test_ids: list[int]) -> Generator[Tuple[str, str], None, None]:
        for idx in test_ids:
            in_path = LocalFS.get_sample_path(website, contest_id, task_id, SampleType.IN, idx)
            out_path = LocalFS.get_sample_path(website, contest_id, task_id, SampleType.OUT, idx)
            yield LocalFS.read(in_path), LocalFS.read(out_path)



from bs4 import BeautifulSoup
import requests
from selenium.webdriver import Chrome
from selenium.webdriver import ChromeOptions


# pylint: disable=too-few-public-methods
class Scraper:

    def __init__(self) -> None:
        options = ChromeOptions()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-gpu')
        options.add_argument('--disable-dev-shm-usage')
        options.add_argument("--window-size=1920,1080")

        self.driver = Chrome(options=options)

    def get(self, url: str) -> BeautifulSoup:
        status_code = requests.get(url, timeout=100).status_code
        if status_code != 200:
            raise ValueError(f'Failed to get {url}: {status_code}')
        page = None
        self.driver.get(url)
        page = self.driver.page_source
        return BeautifulSoup(page, 'lxml')



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
import os

import click
import dotenv

dotenv.load_dotenv()


class Website(Enum):
    ATCODER = 1

    @staticmethod
    def from_string(website: str) -> Website:
        return Website[website.upper()]

    def __str__(self) -> str:
        if self == Website.ATCODER:
            return 'AtCoder'
        raise NotImplementedError()


class SampleType(Enum):
    IN = 1
    OUT = 2

    def __str__(self) -> str:
        return self.name.lower()


@dataclass
class Config:
    WEBSITE_HOST = {Website.ATCODER: 'https://atcoder.jp'}

    LUCY_HOME = os.getenv('LUCY_HOME') or f'{os.getenv("HOME")}/.lucy'
    LUCY_HOME = os.path.abspath(LUCY_HOME)

    SNIPPETS_DIR = f'{LUCY_HOME}/.vscode'
    SNIPPETS_PATH = f'{SNIPPETS_DIR}/cp.code-snippets'

    COMMONS_DIR = os.path.abspath(f'{LUCY_HOME}/common')

    TEMPLATE_PATH = f'{COMMONS_DIR}/base.cpp'

    CLI_WEBSITE_CHOICE = click.Choice(['atcoder'])

    IMPL_MAIN = 'main.cpp'
    IMPL_BIN = 'main'

    SAMPLES_DIR = 'tests'

    # pylint: disable=line-too-long
    DROPBOX_TOKEN = os.getenv(
        'DROPBOX_TOKEN'
    ) or 'sl.B1Rp-zCsGfT8Pd3K_LSV-CduAGyJrBbltWKlly2JvHDzRf3u9AG68md6obRbVyoygMgytF_2L5SYoLijtyQ99qsHH9d9r1SdomNlWZk7uyPTpN4TxhtvnRlbuEfXfpGPF285E61xlXNz'

    ATCODER_TESTCASES_DROPBOX_LINK = 'https://www.dropbox.com/sh/nx3tnilzqz7df8a/AAAYlTq2tiEHl5hsESw6-yfLa?dl=0'
    # pylint: enable=line-too-long



from typing import Any, Generator, Optional, Tuple

import click

from lucy import utils
from lucy.config import Config, Website
from lucy.scraper import Scraper

from lucy.parser_.parser_ import Parser, Task


class AtCoderParser(Parser):

    @staticmethod
    def __parse_samples(task_page: Any) -> Generator[Tuple[str, str], None, None]:
        for input, output in utils.batched(task_page.select('pre[id]'), 2):
            yield input.text, output.text

    def __parse_tasks(self, page: Any) -> list[Task]:
        tasks: list[Task] = []

        tasks_table = page.select('table.table tbody tr')
        for row in tasks_table:
            data = row.find_all('td')
            task_id, task_name = data[0].text, data[1].text

            if self.task_id is not None and task_id != self.task_id:
                continue

            task_path = f'{self.contest_id}_{task_id.lower()}'
            task_page = Scraper().get(f'{self.tasks_page_url}/{task_path}')
            samples = [sample for sample in AtCoderParser.__parse_samples(task_page)]
            click.secho(f'Found {len(samples)} samples for task {task_id}.', fg='green', bold=True)

            task = Task(task_id, task_name, samples)
            tasks.append(task)

        return tasks

    def __init__(self, contest_id: str, task_id: Optional[str]) -> None:
        self.contest_id = contest_id
        self.task_id = task_id

        scraper = Scraper()

        self.tasks_page_url = f'{Config.WEBSITE_HOST[Website.ATCODER]}/contests/{contest_id}/tasks'
        tasks_page = scraper.get(self.tasks_page_url)

        tasks = self.__parse_tasks(tasks_page)

        super().__init__(tasks)



import json
import os
from typing import Generator


def files(path: str) -> Generator[str, None, None]:
    for entry in os.scandir(path):
        if entry.is_file():
            yield entry.path
        elif entry.is_dir():
            yield from files(entry.path)


def generate_snippet(path: str) -> tuple[str, dict[str, str | list[str]]]:
    filename = os.path.basename(path)
    name, ext = filename.split('.')
    scope = ext
    prefix = name
    body: list[str] = []
    with open(path, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.strip('\n')
            if name != 'base':
                if line.startswith(('#include', 'using')):
                    if not line.endswith('//'):
                        continue
                if not body and not line:
                    continue
            body.append(line)
    return name, {
        'scope': scope,
        'prefix': prefix,
        'body': body,
    }


def run(entry_dir: str, snippet_file: str) -> list[str]:
    snippets = {}
    snippet_files = []
    for file in files(entry_dir):
        snippet_files.append(file)
        name, snippet = generate_snippet(file)
        snippets[name] = snippet
    with open(snippet_file, 'w', encoding='utf-8') as f:
        json.dump(snippets, f, indent=2)
    return snippet_files



from typing import Any, Generator


def batched(iterable: list[Any], n: int) -> Generator[tuple[Any, ...], None, None]:
    l = len(iterable)
    for ndx in range(0, l, n):
        yield tuple(iterable[ndx:min(ndx + n, l)])



from typing import Optional

from lucy.config import Website

from lucy.parser_.atcoder import AtCoderParser


class ContestParser:

    def __init__(self, website: Website, contest_id: str, task_id: Optional[str]) -> None:
        self.website = website

        match self.website:
            case Website.ATCODER:
                self.parser = AtCoderParser(contest_id.lower(), task_id)
            case _:
                raise NotImplementedError()
